"""
 Copyright 2024. Aubin Ramon and Pietro Sormanni. CC BY-NC-SA 4.0
 
"""

import pandas as pd
import numpy as np
import os

from joblib.parallel import Parallel, delayed

from .modules.strat_nest import StratifiedNestedCV



def reg_parall_as_job(reg, saved_embed_path, fp_predictions_save, fp_reg_metrics_save, 
                      embeddings, sel_method, strat_method, 
                      random_state, num_outer_fold, num_inner_fold):
    
    if not os.path.exists(fp_predictions_save):
        os.makedirs(fp_predictions_save)
    fp_predictions_df = os.path.join(fp_predictions_save, f"{reg}_pred.csv")

    if not os.path.exists(fp_reg_metrics_save):
        os.makedirs(fp_reg_metrics_save)
    fp_reg_metrics = os.path.join(fp_reg_metrics_save, f"{reg}_perf.csv")

    embed_df = pd.read_pickle(saved_embed_path)
    print(embed_df.columns)

    if not os.path.exists(fp_predictions_df):
        df = embed_df[['id','tm']]
        df.to_csv(fp_predictions_df,index=False,header=True)

    # Data Target Loading
    target = embed_df['tm'].values
    if strat_method == 'none':
        label = embed_df['tm'].values
    elif strat_method == 'loco':
        label = embed_df['clust_kmedoids_blosum'].values
    else:
        label = embed_df[f'{strat_method}'].values

    # Loop over all embeddings
    list_class,list_embed,list_sel_method = [],[],[]

    list_fin_pearson_train, list_fin_pearson_p_train, list_std_pearson_train = [],[],[]
    list_fin_spearman_train, list_fin_spearman_p_train, list_std_spearman_train = [],[],[]
    list_fin_RSME_train, list_std_RSME_train = [],[]
    list_fin_SDR_train, list_std_SDR_train = [],[]

    list_fin_pearson_test, list_fin_pearson_p_test, list_std_pearson_test = [],[],[]
    list_fin_spearman_test, list_fin_spearman_p_test, list_std_spearman_test = [],[],[]
    list_fin_RSME_test, list_std_RSME_test = [],[]
    list_fin_SDR_test, list_std_SDR_test = [],[]

    list_num_seed = []

    for embed_method in embeddings:
        print(f"\n-->> ENCODING sequences with {embed_method} <<--\n")
        data = embed_df[f'{embed_method}'].values

        train_metrics, test_metrics, list_predictions = StratifiedNestedCV(
                                data, target, label,
                                num_outer_fold, num_inner_fold, 
                                random_state, regression=reg, 
                                feature_sel_method=sel_method, 
                                strat_method=strat_method)

        list_class.append(strat_method)
        list_embed.append(embed_method)
        list_sel_method.append(sel_method)

        list_fin_pearson_train.append(train_metrics[0])
        list_fin_pearson_p_train.append(train_metrics[1])
        list_std_pearson_train.append(train_metrics[2])
        list_fin_spearman_train.append(train_metrics[3])
        list_fin_spearman_p_train.append(train_metrics[4])
        list_std_spearman_train.append(train_metrics[5])
        list_fin_RSME_train.append(train_metrics[6])
        list_std_RSME_train.append(train_metrics[7])
        list_fin_SDR_train.append(train_metrics[8])
        list_std_SDR_train.append(train_metrics[9])

        list_fin_pearson_test.append(test_metrics[0])
        list_fin_pearson_p_test.append(test_metrics[1])
        list_std_pearson_test.append(test_metrics[2])
        list_fin_spearman_test.append(test_metrics[3])
        list_fin_spearman_p_test.append(test_metrics[4])
        list_std_spearman_test.append(test_metrics[5])
        list_fin_RSME_test.append(test_metrics[6])
        list_std_RSME_test.append(test_metrics[7])
        list_fin_SDR_test.append(test_metrics[8])
        list_std_SDR_test.append(test_metrics[9])

        list_num_seed.append(len(random_state))

        # Save predictions
        predictions_df = pd.read_csv(fp_predictions_df)
        col = f"{embed_method}_{strat_method}"
        predictions_df[col] = ['_'.join([str(tm) for tm in preds]) for preds in np.array(list_predictions).T]
        predictions_df.to_csv(fp_predictions_df, index=False, header=True)

    # Dictionary of evaluation metrics
    performance_dict = {'strat_method': list_class,
                        'num_seed': list_num_seed, 
                        'embeddings': list_embed,
                        'feature_sel': list_sel_method,
                        
                        'mean_pearson_train': list_fin_pearson_train,
                        'mean_pearson_p_train': list_fin_pearson_p_train,
                        'std_pearson_train': list_std_pearson_train,
                        'mean_pearson_test': list_fin_pearson_test,
                        'mean_pearson_p_test': list_fin_pearson_p_test,
                        'std_pearson_test': list_std_pearson_test,
                                                
                        'mean_spearman_train': list_fin_spearman_train,
                        'mean_spearman_p_train': list_fin_spearman_p_train,
                        'std_spearman_train': list_std_spearman_train,
                        'mean_spearman_test': list_fin_spearman_test,
                        'mean_spearman_p_test': list_fin_spearman_p_test,
                        'std_spearman_test': list_std_spearman_test,
                                                
                        'mean_mae_train': list_fin_RSME_train,
                        'std_mae_train': list_std_RSME_train,
                        'mean_mae_test': list_fin_RSME_test,
                        'std_mae_test': list_std_RSME_test,
                        
                        'mean_sdr_train': list_fin_SDR_train,
                        'std_sdr_train': list_std_SDR_train,
                        'mean_sdr_test': list_fin_SDR_test,
                        'std_sdr_test': list_std_SDR_test
                        }
    
    performance_df = pd.DataFrame(performance_dict)
    print(f"\nPerformance metrics: {performance_df}")

    # Save metrics to csv
    if not os.path.exists(fp_reg_metrics):
        performance_df.to_csv(fp_reg_metrics, mode='w', index=False, header=True)
    else:
        performance_df.to_csv(fp_reg_metrics, mode='a', index=False, header=False)



if __name__ == "__main__":
    
    # File paths
    saved_embed_path = '../../data/database_640.pkl'

    # Folder where to save performances and predictions
    fp_predictions_save = "./saved_results/predictions_exp_blosum_640_mae_no_mean" 
    fp_reg_metrics_save = './saved_results/performances_exp_blosum_640_mae_no_mean'

	# Parameters for trial
    num_outer_fold = 3
    num_inner_fold = 3
    random_state = [0, 42, 100]

    regressions = [
                    'ridge',
        			 'huber',
                	 'GPR',
                	 'SVR',
                	 'elasticnet',
                	 'RF',
                     'LightGBM',
                   ]
    
    embeddings = [
                   'onehot',
                   'esm1b',
                   'esm2_t30',
                   'antiberty',
                   'ablang',
                    'nanobuilder',
                    'vhse',
                 'nanobert',
                  ]
    
    feature_sel_method = 'SelectFromModel' # ['SelectFromModel','PCA', None]
    
    strat_methods = ['expmeth_clust_kmedoids_blosum']

    # Evaluate single regression models, loop over all regressions
    for strat_method in strat_methods:
        Parallel(n_jobs=-1)(delayed(reg_parall_as_job)(reg, saved_embed_path,
                                                    fp_predictions_save, fp_reg_metrics_save,
                                                    embeddings, feature_sel_method, 
                                                    strat_method, random_state,
                                                    num_outer_fold, num_inner_fold) 
                                            for reg in regressions)
    


"""
 Copyright 2024. Aubin Ramon and Pietro Sormanni. CC BY-NC-SA 4.0
"""

import argparse
import sys

from . import predict
from . import train

USAGE = """

%(prog)s <command> [options]

NanoMelt provides one command:
    - predict: predict the apparent melting temperatures of given nanobody sequences
    - train: train NanoMelt with a new input dataset

see also
%(prog)s <command> predict -h
%(prog)s <command> train -h
for additional help

Copyright 2024. Aubin Ramon and Pietro Sormanni. CC BY-NC-SA 4.0
"""


def main():

    if len(sys.argv) == 1:
        empty_parser = argparse.ArgumentParser(
            description="Semi-supervised ensemble model trained to predict nanobody thermostability",
            usage=USAGE
        )
        empty_parser.print_help(sys.stderr)
        sys.exit(1)

    parser = argparse.ArgumentParser(
        description="Semi-supervised ensemble model trained to predict nanobody thermostability",
    )

    subparser = parser.add_subparsers()

    # PREDICT
    predict_parser = subparser.add_parser("predict",
                                        description="Predict nanobody apparent melting temperatures with NanoMelt",
                                        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    predict_parser.add_argument('-i', '--input_filepath_or_seq', help='Filepath to the fasta file .fa to score or directly \
                              a single string sequence', type=str,
                              default='to_score.fa')
    
    predict_parser.add_argument('-o', '--output_savefp', help='Filename of the .csv file to save the predictions in', type=str,
                              default='nanomelt_run.csv')

    predict_parser.add_argument('-align', '--do_align', help='Do the alignment and the cleaning of the given sequences before training. \
                              This step can takes a lot of time if the number of sequences is huge.', action="store_true")
    
    predict_parser.add_argument('-ncpu', '--ncpu', help='If ncpu>1 will parallelise the alignment process', type=int, default=1)

    predict_parser.add_argument('-v', '--verbose', help='Print more details about every step.', action="store_true")

    predict_parser.set_defaults(func=lambda args: predict.run(args))

    # TRAIN
    trainer_parser = subparser.add_parser("train",
                                        description="Train NanoMelt with a new input dataset.",
                                        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    trainer_parser.add_argument('-i', '--input_csv', help='Filepath to the .csv file of the nanobody dataset. \
                                It needs to have at least columns with Id, Sequence, Experimental method \
                             and Measured apparent melting temperature (C).', type=str,
                              default='new_nano_thermostability_dataset.csv')
    
    trainer_parser.add_argument('-od', '--output_dir', help='Filepath to the output dir where to save the saved models', type=str,
                              default='nanomelt_retrain')

    trainer_parser.add_argument('-align', '--do_align', help='Do the alignment and the cleaning of the given sequences before training. \
                              This step can takes a lot of time if the number of sequences is huge.', action="store_true")
    
    trainer_parser.add_argument('-ncpu', '--ncpu', help='If ncpu>1 will parallelise the alignment process', type=int, default=1)

    trainer_parser.set_defaults(func=lambda args: train.run(args))


    args = parser.parse_args()
    args.func(args)



if __name__ == "__main__":
    main()



"""
 Copyright 2024. Aubin Ramon and Pietro Sormanni. CC BY-NC-SA 4.0
 
"""

from sklearn.linear_model import Ridge,ElasticNet,HuberRegressor
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RBF, Matern, WhiteKernel
from sklearn.ensemble import RandomForestRegressor
from sklearn.svm import SVR
from lightgbm import LGBMRegressor



# regression models and params
elasticnet = [ElasticNet(alpha=0.5,l1_ratio=0.1,max_iter=10000,tol=0.01), 
              {'alpha':[0.01,1,10,50,100,200,300,500],
               'l1_ratio':[.1, .5, .7, .9, .95, .99, 1]}]

huber = [HuberRegressor(max_iter=50000, alpha=100, epsilon=5,tol=0.1),
         {'epsilon':[1,1.5,5,10,30,50,100,300,500],
          'alpha':[0.0001,0.01,1,10,50,100,200,300,500]}]

ridge = [Ridge(alpha=200),
         {'alpha':[0.01,1,10,50,100,200,300,500,700,1000]}]

gpr = [GaussianProcessRegressor(random_state=0, n_restarts_optimizer=20, normalize_y=True, alpha=1e-4, 
                                kernel=RBF(length_scale_bounds=(1e-20, 1e5)) + Matern(nu=2.5, length_scale_bounds=(1e-20, 1e10)) + WhiteKernel(noise_level_bounds=(1e-50, 1e10))),
        {'alpha':[0.00001,0.0001,0.001,0.01,1,10,50,100]}]

randomforest = [RandomForestRegressor(max_depth=2, random_state=0),
                {'max_depth':range(1,15),}]

supportvector = [SVR(C=10, epsilon=1e-4),
                 {'kernel':['linear', 'poly', 'rbf', 'sigmoid'],
                  'C':[0.01,0.1,1,10,20,50],
                  'epsilon':[1e-1,1e-3,1e-4,1e-5,1e-6,1e-8]}]

lightgbm = [LGBMRegressor(max_depth=5,path_smooth=5,verbosity=-1),
            {'max_depth':[-1,1,5,10,15,20],
             'extra_trees':['true','false'],
             'path_smooth':[0.0,0.1,0.5,1,5,10,50]}]


# Store regressions in a dict

regression_list = ['elasticnet','RF', 'huber', 'ridge','GPR','SVR','LightGBM']

regression_dict = { 'elasticnet':elasticnet,'RF':randomforest,
                    'huber':huber,'ridge':ridge,'GPR':gpr,
                    'SVR':supportvector,'LightGBM':lightgbm}



"""
 Copyright 2024. Aubin Ramon and Pietro Sormanni. CC BY-NC-SA 4.0

 Generating corresponding embeddings for NanoMelt
"""

import esm
print('Loading ESM models')
esm_models_dict={'esm1b':esm.pretrained.load_model_and_alphabet('esm1b_t33_650M_UR50S'),
                    'esm2_t30':esm.pretrained.load_model_and_alphabet('esm2_t30_150M_UR50D'),
                    }
print('ESM loading complete\n')


import numpy as np
import time
from typing import Tuple

import pandas as pd
from pandas.api.types import CategoricalDtype


import ablang
import torch
import concurrent.futures
from ..modules.alignment.mybio import anarci_alignments_of_Fv_sequences_iter

from ImmuneBuilder import NanoBodyBuilder2
from antiberty import AntiBERTyRunner

from transformers import pipeline, RobertaTokenizer

from functools import wraps
from typing import List
from tqdm import tqdm

def time_it(func):
    @wraps(func)
    def wrapper(*args,**kwargs):
        start = time.time()
        result = func(*args,**kwargs)
        verbose = kwargs.get('verbose', False)
        if verbose:
            print(f'Time taken by {func.__name__} is {time.time() - start}')

        return result
    return wrapper
    
#### SAVED INFO FOR EMBEDDINGS ####
# Alphabet for one hot
alphabet = ['A','C','D','E','F','G','H','I','K','L','M','N','P','Q','R','S','T','V','W','Y','-']

# From reference VHSE-based predictor 10.1371/journal.pone.0074506
vhse_scale_dict = {'A': [0.15, -1.11, -1.35, -0.92, 0.02, -0.91, 0.36, -0.48],
                    'C': [0.18, -1.67, -0.46, -0.21, 0.0, 1.2, -1.61, -0.19],
                    'D': [-1.15, 0.67, -0.41, -0.01, -2.68, 1.31, 0.03, 0.56],
                    'E': [-1.18, 0.4, 0.1, 0.36, -2.16, -0.17, 0.91, 0.02],
                    'F': [1.52, 0.61, 0.96, -0.16, 0.25, 0.28, -1.33, -0.2],
                    'G': [-0.2, -1.53, -2.63, 2.28, -0.53, -1.18, 2.01, -1.34],
                    'H': [-0.43, -0.25, 0.37, 0.19, 0.51, 1.28, 0.93, 0.65],
                    'I': [1.27, -0.14, 0.3, -1.8, 0.3, -1.61, -0.16, -0.13],
                    'K': [-1.17, 0.7, 0.7, 0.8, 1.64, 0.67, 1.63, 0.13],
                    'L': [1.36, 0.07, 0.36, -0.8, 0.22, -1.37, 0.08, -0.62],
                    'M': [1.01, -0.53, 0.43, 0.0, 0.23, 0.1, -0.86, -0.68],
                    'N': [-0.99, 0.0, -0.37, 0.69, -0.55, 0.85, 0.73, -0.8],
                    'P': [0.22, -0.17, -0.5, 0.05, -0.01, -1.34, -0.19, 3.56],
                    'Q': [-0.96, 0.12, 0.18, 0.16, 0.09, 0.42, -0.2, -0.41],
                    'R': [-1.47, 1.45, 1.24, 1.27, 1.55, 1.47, 1.3, 0.83],
                    'S': [-0.67, -0.86, -1.07, -0.41, -0.32, 0.27, -0.64, 0.11],
                    'T': [-0.34, -0.51, -0.55, -1.06, -0.06, -0.01, -0.79, 0.39],
                    'V': [0.76, -0.92, -0.17, -1.91, 0.22, -1.4, -0.24, -0.03],
                    'W': [1.5, 2.06, 1.79, 0.75, 0.75, -0.13, -1.01, -0.85],
                    'Y': [0.61, 1.6, 1.17, 0.73, 0.53, 0.25, -0.96, -0.52],
                    '-': [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
                    }

if torch.cuda.is_available(): print("Transferred ESM model to GPU")
elif torch.backends.mps.is_available(): print("Transferred ESM model to Apple MX GPU")


#### EMBEDDING FUNCTIONS ####
@time_it
def embed_esm_batch(indices: List[int], sequences: List[str], esm_models_dict: dict,
                    esm_model='esm2_t30_150M_UR50D', 
                    batch_size=128, verbose=False) -> Tuple[List[int], np.ndarray]:
    
    
    model, esm_alphabet = esm_models_dict[f'{esm_model}']
    batch_converter = esm_alphabet.get_batch_converter()

    id_layer = model.num_layers

    model.eval()  # disables drop-out for deterministic results
    if torch.cuda.is_available():
        model = model.cuda()
        device = 'cuda'
        
    elif torch.backends.mps.is_available():
        model = model.to('mps')
        device = 'mps'
        
    else:
        device = 'cpu'

    batch_labels, batch_strs, batch_tokens = batch_converter([(str(i), seq) for i, seq in enumerate(sequences)])

    save_seq_encodings = list()
    with torch.no_grad():
        for batch_idx in range(0, len(sequences), batch_size):
            batch_end = min(batch_idx + batch_size, len(sequences))
            toks = batch_tokens[batch_idx:batch_end]
            strs_batch = batch_strs[batch_idx:batch_end]
            # if verbose: print(f"Processing batch {batch_idx + 1} ({len(strs_batch)} sequences)")

            toks = toks.to(device=device, non_blocking=True)
            results = model(toks, repr_layers=[id_layer], return_contacts=False)

            token_representations = results["representations"][id_layer].to(device="cpu")

            representations_per_seq = []
            for i, str_i in enumerate(strs_batch):
                representations_per_seq.append(token_representations[i, 1 : len(str_i) + 1].mean(0))

            seq_coding = np.array([tensor.numpy() for tensor in representations_per_seq], dtype='float64')
            save_seq_encodings.extend(seq_coding)

    return indices, save_seq_encodings, esm_model

@time_it
def embed_ablang_batch(indices: List[int], sequences: List[str], verbose=False) -> Tuple[List[int], np.ndarray]:
    heavy_ablang = ablang.pretrained("heavy") #Automatically uses CUDA if available
    heavy_ablang.freeze()
    seq_coding = np.array(heavy_ablang(sequences, mode='seqcoding'))
    return indices, seq_coding, 'ablang'

@time_it
def embed_nanobuilder_batch(indices: List[int], sequences: List[str], verbose=False) -> Tuple[List[int], List[np.ndarray]]:
    predictor = NanoBodyBuilder2()
    sliced_list_embed = []
    for seq in sequences:
        seq_dict = {'H': seq}
        nanobody = predictor.predict(seq_dict)
        seq_coding = nanobody.encodings[0].mean(0).cpu().numpy()
        sliced_list_embed.append(seq_coding)
    return indices, sliced_list_embed, 'nanobuilder'

@time_it
def embed_vhse_batch(indices: List[int], al_sequences: List[str], verbose=False) -> Tuple[List[int], List[np.ndarray]]:
    list_embedding = []
    for seq in al_sequences:
        seq_coding = np.concatenate([vhse_scale_dict[aa] for aa in seq], axis=None)
        list_embedding.append(seq_coding)
    return indices, list_embedding, 'vhse'


@time_it
def embed_antiberty_batch(indices: List[int], sequences: List[str], verbose=False) -> Tuple[List[int], List[np.ndarray]]:
    antiberty = AntiBERTyRunner() #Automatically uses CUDA if available
    list_tensor = antiberty.embed(sequences)
    list_embedding = [tensor[1:-1].mean(dim=0).cpu().numpy() for tensor in list_tensor]
    return indices, list_embedding, 'antiberty'

@time_it
def embed_nanobert_batch(indices: List[int], sequences: List[str], verbose=False) -> Tuple[List[int], List[np.ndarray]]:
    tokenizer = RobertaTokenizer.from_pretrained("NaturalAntibody/nanoBERT", return_tensors="pt")
    emb = pipeline('feature-extraction', model="NaturalAntibody/nanoBERT", tokenizer=tokenizer)
    list_embedding = [np.mean(emb(seq)[0][1:-1], axis=0) for seq in sequences]
    return indices, list_embedding, 'nanobert'

@time_it
def one_hot_align_batch(indices: List[int], al_sequences: List[str], verbose=False) -> Tuple[List[int], List[np.ndarray]]:
    sliced_list_embed = []
    for seq in al_sequences:
        res_coding = np.array((pd.get_dummies(pd.Series(list(seq)).astype(CategoricalDtype(categories=alphabet))))).astype(float)
        seq_coding = np.concatenate(res_coding, axis=None)
        sliced_list_embed.append(seq_coding)
    return indices, sliced_list_embed, 'onehot'


#### RUN FULL EMBEDDING ####
@time_it
def embed_fasta(seq_records, list_embs: list,
                batch_size: int=128, do_align: bool=True, ncpus: int=8,
                verbose: bool=False) -> pd.DataFrame:
    '''
    Take an input sequences and generate the corresponding embeddings.
    Align sequences with the module in AbNatiV if needed (used for onehot and vhse embeddings).
    Embedding generation is speed up via use of GPU when available and multiple threads. 
    (ESM, Ablang and AntiBerty are the longest ones but they can use GPUs).

    Parameters
    ----------
        - seq_records: list
            List of SeqRecords from the BioPython package. seq = str(record.seq) / id = record.id
        - list_embs: list[str]
            List of embeddings names to compute the embeddings of.
                i.e.: ['esm1b', 'esm2_t30', 'ablang', 'vhse', 'onehot', 'antiberty']
        - batch_size: int
        - do_align: bool
            If True, will align all the sequences using ANARCI (as implemented in AbNatiV).
        - ncpus: int
			Number of CPUs to use to paralelise the alignement and the embedding step of the dataset.
        - verbose: bool
			If True, will print more information about every step.

    Returns
    ----------
        - A dataframe with the saved sequences and embeddings. 
    '''

    # Load the data and align sequences if needed
    if type(seq_records) is not list and type(seq_records) is not tuple :
        seq_records=[seq_records]
    
    if type(seq_records[0]) is str :
        list_seq = [str(rec) for rec in seq_records]
        list_id = [str(j) for j in range(len(seq_records))]
    else :
        list_seq = [str(rec.seq) for rec in seq_records]
        list_id = [rec.id for rec in seq_records]

    if do_align:
        start_al = time.process_time()
        VH,_,_,_,_ = anarci_alignments_of_Fv_sequences_iter(seq_records, isVHH=True,
                                                       del_cyst_misalign=False, check_AHo_CDR_gaps=True, 
                                                       run_parallel=ncpus, verbose=verbose)
        recs = VH.to_recs()
        list_id = [rec.id for rec in recs]
        list_al_seq = [str(rec.seq) for rec in recs]
        list_seq = [seq.replace('-','') for seq in list_al_seq]
        if verbose: print('Time making alignment: ', time.process_time() - start_al)
    else:
        list_al_seq = list_seq
        list_seq = [seq.replace('-','') for seq in list_seq]

    with concurrent.futures.ThreadPoolExecutor() as executor:
        futures1 = []

        if 'esm1b' in list_embs: futures1.append(executor.submit(embed_esm_batch, range(len(list_seq)), list_seq, esm_models_dict, 'esm1b', batch_size, verbose))
        if 'esm2_t30' in list_embs: futures1.append(executor.submit(embed_esm_batch, range(len(list_seq)), list_seq, esm_models_dict, 'esm2_t30', batch_size, verbose))

        all_esm1_embs, all_esm2_embs = dict(), dict()
        all_ablang_embs, all_nanobuilder_embs, all_antibery_embs = dict(), dict(), dict()
        all_nanobert_embs, all_vhse_embs, all_onehot_embs = dict(), dict(), dict() 

        # Futures2 not to send as threads as needed batches at the same time 
        if verbose: print('\nGenerating embeddings over batches')
        for i in tqdm(range(0, len(list_seq), batch_size), disable=True):
            batch_indices = list(range(i, min(i+batch_size, len(list_seq))))
            batch_seqs = list_seq[i:i+batch_size]
            batch_al_seqs = list_al_seq[i:i+batch_size]

            futures2 = []
            if 'ablang' in list_embs: futures2.append(executor.submit(embed_ablang_batch, batch_indices, batch_seqs, verbose))
            if 'vhse' in list_embs: futures2.append(executor.submit(embed_vhse_batch, batch_indices, batch_al_seqs, verbose))
            if 'antiberty' in list_embs: futures2.append(executor.submit(embed_antiberty_batch, batch_indices, batch_seqs, verbose))
            if 'onehot' in list_embs: futures2.append(executor.submit(one_hot_align_batch, batch_indices, batch_al_seqs, verbose))
            if 'nanobuilder' in list_embs: futures2.append(executor.submit(embed_nanobuilder_batch, batch_indices, batch_seqs, verbose))
            if 'nanobert' in list_embs: futures2.append(executor.submit(embed_nanobert_batch, batch_indices, batch_seqs, verbose))

            concurrent.futures.wait(futures2)

            # Collect results for the current batch
            for future in futures2:
                indices, result, embedding_type = future.result()
                if embedding_type == 'ablang':
                    all_ablang_embs.update(dict(zip(indices, result)))
                elif embedding_type == 'vhse':
                    all_vhse_embs.update(dict(zip(indices, result)))
                elif embedding_type == 'antiberty':
                    all_antibery_embs.update(dict(zip(indices, result)))
                elif embedding_type == 'onehot':
                    all_onehot_embs.update(dict(zip(indices, result)))
                elif embedding_type == 'nanobert':
                    all_nanobert_embs.update(dict(zip(indices, result)))
                elif embedding_type == 'nanobuilder':
                    all_nanobuilder_embs.update(dict(zip(indices, result)))

        for future in concurrent.futures.as_completed(futures1):
            indices, result, embedding_type = future.result()
            if embedding_type == 'esm1b':
                all_esm1_embs.update(dict(zip(indices,result)))
            elif embedding_type == 'esm2_t30':
                all_esm2_embs.update(dict(zip(indices,result)))

    df_embed = pd.DataFrame({
        'id': list_id,
        'al_seq': list_al_seq,
        'seq': list_seq})
    
    if 'esm1b' in list_embs: df_embed['esm1b'] = [all_esm1_embs[i] for i in range(len(list_seq))]
    if 'esm2_t30' in list_embs: df_embed['esm2_t30'] = [all_esm2_embs[i] for i in range(len(list_seq))]
    if 'ablang' in list_embs: df_embed['ablang'] = [all_ablang_embs[i] for i in range(len(list_seq))]
    if 'vhse' in list_embs: df_embed['vhse'] = [all_vhse_embs[i] for i in range(len(list_seq))]
    if 'antiberty' in list_embs: df_embed['antiberty'] = [all_antibery_embs[i] for i in range(len(list_seq))]
    if 'onehot' in list_embs: df_embed['onehot'] = [all_onehot_embs[i] for i in range(len(list_seq))]
    if 'nanobuilder' in list_embs: df_embed['nanobuilder'] = [all_nanobuilder_embs[i] for i in range(len(list_seq))]
    if 'nanobert' in list_embs: df_embed['nanobert'] = [all_nanobert_embs[i] for i in range(len(list_seq))]

    return df_embed




"""
 Copyright 2024. Aubin Ramon and Pietro Sormanni. CC BY-NC-SA 4.0
 
"""

from .alignment import mybio
from .alignment.blossum import distance_blosum_normalised_gap_to_zero, distance_blosum_normalised


from sklearn_extra.cluster import KMedoids
from scipy.spatial.distance import squareform

import numpy as np
from kneed import KneeLocator




def kmedoids_clustering(dist_matrix: np.array, max_nb_clusters: int) -> list:
    '''Clustering using the KMedoids algorithm. It will iterate up to max_nb_clusters number
    of clusters. It applied the elbow method on the loss ot find the optimal number of clusters. 

    Parameters
    ----------
        - dist_matrix: np.array
            A (N,N) matrix that represent the distance among and betweenN sequences.
        - max_nb_clusters: int
            Maximun number of clusters to try out.
    Returns
    -------
        - found_labels: list
            List of label for each element (in the same order) of dist_matrix.
                e.g., [1,1,1,3,1,4,4,2,1,2,2]

    '''
    if len(dist_matrix) < max_nb_clusters:
        max_nb_clusters = len(dist_matrix)

    sse = []
    for k in range(1,max_nb_clusters+1):
        c = KMedoids(n_clusters=k, metric='precomputed',
                         init='k-medoids++', method='pam',
                         random_state=0)
        c.fit(dist_matrix)
        sse.append(c.inertia_)  # within-cluster sum of distances

    kl = KneeLocator(range(1, max_nb_clusters+1), sse, curve="convex", direction="decreasing")
    best_nb_clust = kl.elbow
    if best_nb_clust is None:
        best_nb_clust = max_nb_clusters
        print(f'--> Could not find Elbos (max_cluster {max_nb_clusters} might be too low, so take max_cluster as optimal nb of clusters.')
    else: print(f'--> SSE elbow computed as {best_nb_clust} by kneed')


    final = KMedoids(n_clusters=int(best_nb_clust), metric='precomputed',
                     init='k-medoids++', method='pam',
                     random_state=0)
    final.fit(dist_matrix)
    founds_labels = final.labels_

    print(f'\n## Kmedoids clustering ##')

    kl = KneeLocator(range(1, max_nb_clusters+1), sse, curve="convex", direction="decreasing")
    print(f'--> SSE elbow computed as {kl.elbow} by kneed (out of {max_nb_clusters} tested)')

    return founds_labels
    


def get_k_medoids_clusters(aligned_sequences: list) -> list:
    ''' 
    Take aligned sequences as an input, compute BLOSUM matrix
    ditances, and cluster in using ELBOW method with k-medoids. 

    Returns a list of label clusters.
    '''
    # Compute distance matrices
    mat_for_dist = distance_blosum_normalised
    distance_mat = mybio.distance_matrix_from_aligned_sequences(aligned_sequences, matrix_for_distance=mat_for_dist)
    matrix_dissim = squareform(distance_mat)

    # Normalised nb of mutations with the average AHo length 
    norm_matrix_mut = np.zeros((distance_mat.shape))

    for i, seq_nb_mutsin in enumerate(distance_mat):
        for j, nb_mut in enumerate(seq_nb_mutsin):
            l = (len(aligned_sequences[i].replace('-',''))+len(aligned_sequences[j].replace('-','')))/2
            norm_matrix_mut[i,j] = nb_mut/l

    labels_kmedoids_blosum = kmedoids_clustering(norm_matrix_mut, max_nb_clusters=40)

    return labels_kmedoids_blosum


"""
 Copyright 2024. Aubin Ramon and Pietro Sormanni. CC BY-NC-SA 4.0
 
"""

import numpy as np
import pandas as pd
import os

from .modules.ensemble import TopPred
from .modules.utils import perf_metrix

if __name__ == "__main__":

    # File paths
    saved_embed_path = "../../data/database_640.pkl"
    strat_method = 'expmeth_clust_kmedoids_blosum' # 'expmeth_clust_kmedoids_blosum' 'exp_method'

    perf_csv_dir = "./saved_results/performances_exp_blosum_640_mae_no_mean"
    saved_predictions_path = "./saved_results/predictions_exp_blosum_640_mae_no_mean"

    metric = "spearman"

    # Choose diverse embedding or regression
    
    do_diverse_sorting = False
    diverse_emb = False
    diverse_reg = True

    if not do_diverse_sorting: id = 'true_top'
    elif diverse_emb: id = 'div_emb'
    else: id = 'div_reg'

    if not os.path.exists(perf_csv_dir):
        os.makedirs(perf_csv_dir)
    mean_metrics_csv = os.path.join(perf_csv_dir, f"ens_no_mean_{metric}_ave_{id}.csv")

    if not os.path.exists(saved_predictions_path):
        os.makedirs(saved_predictions_path)
    mean_pred_csv = os.path.join(saved_predictions_path, f"ens_no_mean_{metric}_ave_{id}.csv")

    if os.path.exists(mean_metrics_csv): os.remove(mean_metrics_csv)
    if os.path.exists(mean_pred_csv): os.remove(mean_pred_csv)

    embed_df = pd.read_pickle(saved_embed_path)



    # Number of top ranks to average
    num_to_ave = 8

    # parameters
    
    num_tm_list = np.arange(1,num_to_ave+1,1)

    seeds = [0,42,100]

    for num_tm in num_tm_list:
        saved_pearson, saved_spearman, saved_mae, saved_sdr = list(), list(), list(), list()
        list_preds_per_seed = list()
        for seed_idx in range(len(seeds)):
            trial = f"ave_top_{num_tm}"

            predictions = TopPred(top_ranks=num_tm,
                                    metric=metric,
                                    strat_method=strat_method,
                                    fp_database = saved_embed_path, 
                                    perf_csv_dir = perf_csv_dir,
                                    saved_predictions_path = saved_predictions_path, 
                                    diverse_emb=diverse_emb,
                                    diverse_reg=diverse_reg,
                                    do_diverse_sorting=do_diverse_sorting,
                                    seed_idx=seed_idx
                                    )

            ave_pred = [np.mean(i) for i in predictions]
            list_preds_per_seed.append(ave_pred)

            true_tm = embed_df['tm'].values
            id = embed_df['id'].values

            pearson, _, spearman, _, mae, sdr  = perf_metrix(ave_pred,true_tm)
            saved_pearson.append(pearson)
            saved_spearman.append(spearman)
            saved_mae.append(mae)
            saved_sdr.append(sdr)

        mean_pearson, std_pearson = np.mean(saved_pearson), np.std(saved_pearson)
        mean_spearman, std_spearman = np.mean(saved_spearman), np.std(saved_spearman)
        mean_mae, std_mae = np.mean(mae), np.std(mae)
        mean_sdr, std_sdr = np.mean(saved_sdr), np.std(saved_sdr)

        # Save results to csv
        performance_df = {'method': [trial], 'pearson': [mean_pearson], 'std_pearson': [std_pearson], 
                          'spearman': [mean_spearman], 'std_spearman': [std_spearman], 
                          'mae': [mean_mae], 'std_mae': [std_mae],
                          'sdr': [mean_sdr], 'std_sdr': [std_sdr]}
        
        performance_df = pd.DataFrame(performance_df)
        if not os.path.exists(mean_metrics_csv):
            performance_df.to_csv(mean_metrics_csv, mode='w', index=False, header=True)
        else:
            performance_df.to_csv(mean_metrics_csv, mode='a', index=False, header=False)
        
        if not os.path.exists(mean_pred_csv):
            pred_df = {'id': id, 'tm': true_tm, f'{trial}': ['_'.join([str(tm) for tm in preds]) for preds in np.array(list_preds_per_seed).T]}
            pred_df = pd.DataFrame(pred_df)
            pred_df.to_csv(mean_pred_csv, index=False)
        else:
            pred_df = pd.read_csv(mean_pred_csv)
            pred_df[f'{trial}'] = ['_'.join([str(tm) for tm in preds]) for preds in np.array(list_preds_per_seed).T]
            pred_df.to_csv(mean_pred_csv, index=False)
